@@ -9,12 +9,19 @@ void ParticlesMonitor_track_local_particle(ParticlesMonitorData el,
     int64_t const stop_at_turn = ParticlesMonitorData_get_stop_at_turn(el);
     int64_t const part_id_start = ParticlesMonitorData_get_part_id_start(el);
     int64_t const part_id_end= ParticlesMonitorData_get_part_id_end(el);
+    int64_t const ebe_mode = ParticlesMonitorData_get_ebe_mode(el);
     ParticlesData data = ParticlesMonitorData_getp_data(el);
 
     int64_t n_turns_record = stop_at_turn - start_at_turn;
 
     //start_per_particle_block (part0->part)
-	int64_t const at_turn = LocalParticle_get_at_turn(part);
+	int64_t at_turn;
+	if (ebe_mode){
+		at_turn = LocalParticle_get_at_element(part);
+	}
+	else{
+		at_turn = LocalParticle_get_at_turn(part);
+	}
 	if (at_turn>=start_at_turn && at_turn<stop_at_turn){
 	    int64_t const particle_id = LocalParticle_get_particle_id(part);
 	    if (particle_id<part_id_end && particle_id>=part_id_start){
